//*****************************************************/
#define LED1_ON PORTD.6=0
#define LED1_OFF PORTD.6=1

#pragma promotechar+
#pragma uchar+
#pragma regalloc-
#pragma optsize+

#include <delay.h>
#include <cvmegaload.h> //contains defines for DeviceID FlashSize BootSize PageSize AddressLshift

register unsigned int Pagedata @2; //program data to be written from this and read back for checking
register unsigned int PageAddress @4; //address of the page
register unsigned int CurrentAddress @6; //address of the current data -  PageAddress + loop counter
register char inchar @8; //data received from RS232 
register char spmcrval @10; //value to write to SPM control register 
register unsigned int i @11;   //loop counter
register unsigned int j @13;  //loop counter  
unsigned int Checkdata ; //compared with Pagedata for checking
char PageBuffer[PageByte]; //buffer for data to be written   

//*********************************************************************************************************
char GPS_init[] = {'$','P','S','R','F','1','0','3',',','3',',','0',',','0',',','1','*','2','7',13,10,00,00,00,
'$','P','S','R','F','1','0','0',',','1',',','3','8','4','0','0',',','8',',','1',',','0','*','3','D',13,10};

//"$PSRF103,3,0,0,1*27$PSRF100,1,38400,8,1,0*3D"        GPS
//***********************************************************************************************************
// Get a character from the USART1 Receiver
char getchar1(void)
{
char status,data;
while (1)
      {
      while (((status=UCSR1A) & 128)==0);
      data=UDR1;
      if ((status & (28))==0)
         return data;
      };
}
// Write a character to the USART1 Transmitter
void putchar1(char c)
{
while ((UCSR1A & 32)==0);
UDR1=c;
}


char GetPage(void)
{
char LocalCheckSum = 0;
char CheckSum = 0;
for (j=0; j<PageByte; j++)
  {
  PageBuffer[j]=getchar1();
  LocalCheckSum += PageBuffer[j];
  }
CheckSum = getchar1();  
if(LocalCheckSum==CheckSum) return 1;
else return 0;
}

char CheckFlash(void)
{
//After the data has been written to flash it is read back and compared to the original
for (j=0; j<PageByte; j+=2)
  {
  CurrentAddress=PageAddress+j; 
  #asm
  movw r30, r6       ;//move  CurrentAddress to Z pointer  
  elpm r2, Z+         ;//read LSB
  elpm r3, Z           ;//read MSB    
  #endasm    
  Checkdata=PageBuffer[j] +(PageBuffer[j+1]<<8);
  if(Pagedata != Checkdata) return 0;
  }
return 1;
}  

void ExecCode(void)
{
RAMPZ =  0;  
MCUCR = 0x01;	       // Enable interrupt vector select
MCUCR = 0x00;	       // Move interrupt vector to flash
#asm("jmp 0x00"); // Run application code   
}

void BootLoad(void)
{ 
// Send chip data to the programming software so that it knows
// how to format transmissions
putchar1(DeviceID); 
putchar1(FlashSize);
putchar1(BootSize);
putchar1(PageSize);  
// "!" means all ok and send the next data if there is more
putchar1('!');
while(1)
  {
  PageAddress = (unsigned int)getchar1() << 8;  // Receive PageAddress high byte
  PageAddress += getchar1();   // Add PageAddress low byte
  if(PageAddress==0xffff) ExecCode(); // The windows program sends this value when finished  
  if(PageAddress>>8) RAMPZ=1;
  else RAMPZ=0;  
  PageAddress = PageAddress << AddressLshift; //essentially the same as multiply by PageSize
  if (GetPage()) //receive one page of data followed by a checksum byte and verify data
    {
    for(i=0; i<PageByte; i+=2) //fill temporary buffer in 2 byte chunks from PageBuffer           
      { 
      Pagedata=PageBuffer[i]+(PageBuffer[i+1]<<8);
      while(SPMCR & 1); //wait for spm complete
      CurrentAddress=PageAddress+i; 
      spmcrval=1;
      #asm 
      movw r30, r6    ;//move CurrentAddress to Z pointer   
      mov r1, r3        ;//move Pagedata MSB reg 1
      mov r0, r2        ;//move Pagedata LSB reg 1  
      sts SpmcrAddr, r10   ;//move spmcrval to SPM control register
      spm                ;//store program memory
      #endasm
      }    
    while (SPMCR & 1);  //wait for spm complete
    spmcrval=3;        //erase page
    #asm 
    movw r30, r4       ;//move PageAddress to Z pointer
    sts SpmcrAddr, r10    ;//move spmcrval to SPM control register              
    spm                 ;//erase page
    #endasm
      
    while (SPMCR & 1); //wait for spm complete
    spmcrval=5;        //write page
    #asm 
    movw r30, r4       ;//move PageAddress to Z pointer
    sts SpmcrAddr, r10    ;//move spmcrval to SPM control register              
    spm                 ;//write page
    #endasm

    while (SPMCR & 1);  //wait for spm complete
    spmcrval=0x11;   //enableRWW  see mega8 datasheet for explanation
     // P. 212 Section "Prevent reading the RWW section
     // during self-programming
    #asm 
    sts SpmcrAddr, r10   ;//move spmcrval to SPMCR              
    spm   
    #endasm
    if(CheckFlash()) putchar1('!');  //all ok, send next page
    else putchar1('@'); //there was an error, resend page
    }  //end if (GetPage()) 
  else putchar1('@');  //there was an error ,resend page
  }
}

void main(void)
{    
// Crystal Oscillator division factor: 1
#pragma optsize-
CLKPR=0x80;
CLKPR=0x00;
#ifdef _OPTIMIZE_SIZE_
#pragma optsize+
#endif

// Port D initialization
// Func7=Out Func6=Out Func5=In Func4=In Func3=Out Func2=In Func1=In Func0=In 
// State7=0 State6=0 State5=P State4=P State3=0 State2=P State1=P State0=P 
PORTD=0x37;
DDRD=0xC8;

// USART1 initialization
// Communication Parameters: 8 Data, 1 Stop, No Parity
// USART1 Receiver: On
// USART1 Transmitter: On
// USART1 Mode: Asynchronous
// USART1 Baud Rate: 38400
UCSR1A=0x00;
UCSR1B=0x18;
UCSR1C=0x06;
UBRR1H=0x00;
UBRR1L=0x20;

putchar1('>'); //I'm here, talk to me


for(i=0; i<10; i++)
  {
  if(i & 1) LED1_ON; else LED1_OFF; 
  if(UCSR1A & 0x80)
    {
    inchar=getchar1();
    if(inchar=='<') { BootLoad(); break; }
    }
  delay_ms(50);
 }
//*********************************************************** 
UBRR1H=0x01;
UBRR1L=0x03;  //     4800
//UBRR1H=0x00;
//UBRR1L=0x81;  //     9600
delay_ms(150) ;
for(i=0; i<52; i++)
 {   
  putchar1(GPS_init[i]);  //     GPS
 }
//********************************************************************
ExecCode();  
}
